"""
This module contains the project meta-data.
"""

__author__ = "Caleb P. Burns"
__copyright__ = "Copyright © 2013-2026 Caleb P. Burns"
__credits__ = [
	"Hong Minhee <https://github.com/dahlia>",
	"Brandon High <https://github.com/highb>",
	"029xue <https://github.com/029xue>",
	"Michael Huynh <https://github.com/mikexstudios>",
	"Nick Humrich <https://github.com/nhumrich>",
	"David Fraser <https://github.com/davidfraser>",
	"Charles Samborski <https://github.com/demurgos>",
	"George Hickman <https://github.com/ghickman>",
	"Vincent Driessen <https://github.com/nvie>",
	"Adrien Vergé <https://github.com/adrienverge>",
	"Anders Blomdell <https://github.com/AndersBlomdell>",
	"Xavier Thomas <https://github.com/thmxv>",
	"Wim Jeantine-Glenn <https://github.com/wimglenn>",
	"Hugo van Kemenade <https://github.com/hugovk>",
	"Dan Cecile <https://github.com/dcecile>",
	"MrOutis <https://github.com/mroutis>",
	"Jon Dufresne <https://github.com/jdufresne>",
	"Greg Roodt <https://github.com/groodt>",
	"Florin T. <https://github.com/ftrofin>",
	"Ben Felder <https://github.com/pykong>",
	"Nicholas Hollander <https://github.com/nhhollander>",
	"KOLANICH <https://github.com/KOLANICH>",
	"Jon Hays <https://github.com/JonjonHays>",
	"Isaac0616 <https://github.com/Isaac0616>",
	"Sebastiaan Zeeff <https://github.com/SebastiaanZ>",
	"Roel Adriaans <https://github.com/RoelAdriaans>",
	"Ravi Selker <https://github.com/raviselker>",
	"Johan Vergeer <https://github.com/johanvergeer>",
	"danjer <https://github.com/danjer>",
	"Jan-Hein Bührman <https://github.com/jhbuhrman>",
	"Wim-Peter Dirks <https://github.com/WPDOrdina>",
	"Karthikeyan Singaravelan <https://github.com/tirkarthi>",
	"John Vandenberg <https://github.com/jayvdb>",
	"John T. Wodder II <https://github.com/jwodder>",
	"Tomasz Kłoczko <https://github.com/kloczek>",
	"Oren <https://github.com/orens>",
	"SP Mohanty <https://github.com/spMohanty>",
	"Richard Si <https://github.com/ichard26>",
	"Jakub Kuczys <https://github.com/jack1142>",
	"Michał Górny <https://github.com/mgorny>",
	"Bartłomiej Żak <https://github.com/bzakdd>",
	"Matthias <https://github.com/haimat>",
	"Avasam <https://github.com/Avasam>",
	"Anıl Karagenç <https://github.com/karagenc>",
	"Yannic Schröder <https://github.com/yschroeder>",
	"axesider <https://github.com/axesider>",
	"TomRuk <https://github.com/tomruk>",
	"Oleh Prypin <https://github.com/oprypin>",
	"Lumina <https://github.com/lumina37>",
	"Kurt McKee <https://github.com/kurtmckee>",
	"Dobatymo <https://github.com/Dobatymo>",
	"Tomoki Nakamaru <https://github.com/tomokinakamaru>",
	"Sebastien Eskenazi <https://github.com/SebastienEske>",
	"Bar Vered <https://github.com/barv-jfrog>",
	"Tzach Shabtay <https://github.com/tzachshabtay>",
	"Adam Dangoor <https://github.com/adamtheturtle>",
	"Marcel Telka <https://github.com/mtelka>",
]
__license__ = "MPL 2.0"
