from typing import Final

IMPORT_MAPPING: Final[dict[str, str]]
NAME_MAPPING: Final[dict[tuple[str, str], tuple[str, str]]]
PYTHON2_EXCEPTIONS: Final[tuple[str, ...]]
MULTIPROCESSING_EXCEPTIONS: Final[tuple[str, ...]]
REVERSE_IMPORT_MAPPING: Final[dict[str, str]]
REVERSE_NAME_MAPPING: Final[dict[tuple[str, str], tuple[str, str]]]
PYTHON3_OSERROR_EXCEPTIONS: Final[tuple[str, ...]]
PYTHON3_IMPORTERROR_EXCEPTIONS: Final[tuple[str, ...]]
